/* io.c */
#include <stdio.h>
#include <stdlib.h>
#include "cfp.h"

extern char domainFile[];
extern int No_Features, No_Classes;
extern float delta;
extern float weight[];
extern float gen_limit[];
extern int linear[];
extern int className[], classIndex[];
extern SEGMENT *Partitions[];
extern int LEARN_WEIGHTS;
extern FILE *logf;

error_exit (message1, message2)
char *message1, *message2;
{
  fprintf(stderr, "%s %s\n", message1, message2);
  exit(1);
}

separator(ch)
char ch;
{
  if (ch == ' ' || ch == ',' || ch == '\t')
    return(1);
  else
    return(0);
}

char *read_next_value(into, from)
char *into, *from;
{
  int j;

  /* skip separators */
  while (separator(*from) && *from != '\n') from++;
  j=0; 
  while (!separator(*from) && *from != '\n') {
    into[j++]=*from; 
    from++; 
  }
  into[j]='\0';			/* into contains a value as a string */
  return from;
}

void read_an_instance(fname, i, line)
char *fname;
INSTANCE *i;
char *line;
{
  char value_in_string[MAX_LINE];
  int v;
  FEATURE f=0;
  CLASS c;

  for (v = 0;  v < No_Features; v++) {
    line = read_next_value(value_in_string, line);
    if (value_in_string[0] == '?')
      i->status[f++]=UNKNOWN;
    else { 
      sscanf(value_in_string, "%f", &(i->value[f]));
      i->status[f++]=KNOWN;
    }
  }
  line = read_next_value(value_in_string, line);
  if (value_in_string[0] == '?') 
    error_exit ("Unknown (?) class value in", fname);

  sscanf(value_in_string, "%d", &c);
  if (classIndex[c] == 0) 
    error_exit ("Strange class name in", fname);
  i->class = classIndex[c];
}

void read_data(fname, Data, size)
char *fname;
INSTANCE *Data;
int *size;
{
  FILE *fp;
  int n;
  char line[MAX_LINE];

  fp=fopen(fname, "r");
  if (fp==NULL) 
    error_exit ("cannot open data file", fname);
  
  for (n=0; fgets(line, MAX_LINE, fp) != 0; n++)
    read_an_instance(fname, &Data[n], line);

  *size = n;
  fclose(fp);
}

void show_instance(inst) 
INSTANCE inst;
{
  FEATURE f;

  for (f = 0; f < No_Features; f++)
    if (inst.status[f] == KNOWN) 
      fprintf(logf, "%6g ", inst.value[f]);
    else
      fprintf(logf, "%?????? ");
  fprintf(logf, ": %d\n", className[inst.class]);
}
  
void show_data(Data, size)  /* for diagnostic purposes */
INSTANCE *Data;
int size;
{
  int i;

  for (i = 0; i < size; i++) 
    show_instance(Data[i]);
}

void show_weights()
{
  FEATURE f;
  
  fprintf (logf, "Weigths are: ");
  for (f=0; f < No_Features; f++)
    fprintf (logf, "w[%d]: %g ", f, weight[f]);
  fprintf (logf, "\n");
}

void read_domain(domainFile)   /* read domain dependent parameters */
char *domainFile;
{
  FILE *fp;
  int f;
  char Line[MAX_LINE], str[MAX_LINE], *line;

  fp=fopen(domainFile,"r");
  if (fp==NULL)
    error_exit(domainFile, "file is not found...");

  while (fgets(Line,MAX_LINE,fp) != NULL) {
    if (sscanf(Line,"%s",str) <1) continue; /* ignore empty line */
    if (Line[0]==';') continue;	/* ignore comment line */

    line = read_next_value (str, Line);

    if (strcasecmp(str,"Features")==0) {
      No_Features = 0;
      while ((line=read_next_value(str,line)) && (*str != '\0')) {
	if (No_Features == MAX_FEATURES) 
	  error_exit ("No_Features has now reached to MAX_FEATURES","");
	if (*str == 'l')
	  linear[No_Features++] = TRUE;
	else
	  linear[No_Features++] = FALSE;
      }
    }
    else if (strcasecmp(str,"Classes")==0) {
      No_Classes = 0;
      while ((line=read_next_value(str,line)) && (*str != '\0')) {
	if (No_Classes == MAX_CLASSES)
	  error_exit ("No_Classes has now reached to MAX_CLASSES","");
	className[++No_Classes] = atoi(str);
	classIndex[className[No_Classes]] = No_Classes;
      }
    }
    else if (strcasecmp(str,"delta")==0)  {
      if (sscanf(line,"%f", &delta) <1)
	error_exit(domainFile, "must specify WeightAdjustment rate");
      LEARN_WEIGHTS = TRUE;
    }
    else if (strcasecmp(str,"Weights")==0) {
      for (f=0; f<No_Features; f++) {
	line=read_next_value(str,line);
	if (sscanf(str,"%f", &weight[f]) <1)
	  error_exit(domainFile, "must specify weights");
      }
      LEARN_WEIGHTS = FALSE;
    }
    else if (strcasecmp(str,"Limits")==0) {
      for (f=0; f<No_Features; f++) {
	line=read_next_value(str,line);
	if (sscanf(str,"%f", &gen_limit[f]) <1)
	  error_exit(domainFile, "must specify generalization Limits");
      }
    }
  }
  fclose(fp);
}

void show_partitions ()
{
  FEATURE f;
  SEGMENT *s;

  fprintf(logf, "Partitions:\n");
  for (f=0; f < No_Features ; f++) {
    fprintf(logf, "f[%d]: ", f);
    for (s = Partitions[f]; s != NULL; s = s->right) 
      fprintf(logf, "[%g,%g]:%d:%g ", s->lower,s->upper, s->class, s->repr);
    fprintf(logf, "\n");
  }
}
