/* cfp.c: Classifier with Feature Partitioning */

#include <stdio.h>
#include <string.h>
#include "cfp.h"

extern int No_Features;
extern float gen_limit[MAX_FEATURES];

SEGMENT *Partitions[MAX_FEATURES];

INSTANCE trainData[MAX_DATA], testData[MAX_DATA];

void initialize_partitions ()
{
  FEATURE f;
  SEGMENT *new;

  for (f=0; f < No_Features; f++) {
    new = (SEGMENT *)calloc(1,sizeof(SEGMENT));
    new->lower = -INFINITE;
    new->upper =  INFINITE;
    new->repr = 0.0;
    new->left = new->right = NULL;
    Partitions[f] = new;
  }
}

SEGMENT *search (f, value)
FEATURE f;
VALUE value;
{
  SEGMENT *seg;

  for (seg = Partitions[f]; value > seg->upper; seg = seg->right);
  return (seg);
}

void update_point (seg, value, class)
SEGMENT *seg;
VALUE value;
CLASS class;
{  
SEGMENT *s, *new_point;

  for (s=seg; (s->right !=NULL && s->lower<=value); s= s->right) {
      if (s->class==class) {
         s->repr++;
         return;
      }
  }
  new_point = (SEGMENT *)calloc(1,sizeof(SEGMENT));
  new_point->right=seg->right;
  if (seg->right) 
     seg->right->left = new_point;
  new_point->left=seg;
  new_point->upper=new_point->lower=value;
  new_point->repr=1.0;
  new_point->class=class;
  seg->right=new_point;
}

void subdivide_range(seg, value, class)
SEGMENT *seg;
VALUE value;
CLASS class;
{
  VALUE range;
  SEGMENT *new_point, *new_range;

  range = seg->upper - seg->lower;
  new_point = (SEGMENT *)calloc(1,sizeof(SEGMENT));
  new_range = (SEGMENT *)calloc(1,sizeof(SEGMENT));
  new_point->left = seg;
  new_point->right = new_range;
  new_range->left = new_point;
  new_range->right = seg->right;
  if (seg->right != NULL)
    (seg->right)->left = new_range;
  seg->right = new_point;

  new_point->upper = new_point->lower = value;
  new_point->class = class;
  new_point->repr = 1.0;

  new_range->lower = value;
  new_range->upper = seg->upper;
  new_range->class = seg->class;
  new_range->repr = seg->repr * (new_range->upper - new_range->lower) / range;

  seg->upper = value;
  seg->repr = seg->repr - new_range->repr;
}

/*
   head is on left of the node 
   node->class = UNKNOWN
   head->class = examples[E].class
*/
generalize_left(node,head,value)
SEGMENT *node,*head;
VALUE value;
{

  if (node->left == head) {  /* if head is next to the node */
    node->lower=value;
    head->upper=value;
    head->repr++;
    return(1);
  }
  /* there are/is node(s) between head and the node */
  if (head->left !=NULL)
     head->left->right=head->right;
  if (head->right !=NULL)
     head->right->left=head->left;
  head->right=node;
  head->left=node->left;
  if (node->left !=NULL)
     node->left->right=head; 
  node->left=head;
  head->upper=value;
  node->lower=value;
  head->repr++;
  return(1);
}

/* 
   head is on the right of the node 
   node->class = UNKNOWN
   head->class = examples[E].class
*/
generalize_right(node,head,value)
SEGMENT *node,*head;
VALUE value;
{
  if (node->right == head) {	/* head is next to node */
    head->lower=value;
    head->repr++;
    node->upper=value;
    return(1);
  }
  /* there are/is node(s) between node and head(on right) */
  if (head->left !=NULL)
    head->left->right=head->right;
  if (head->right !=NULL)
    head->right->left=head->left;
  head->right=node->right;
  head->left=node;
  if (node->right !=NULL)
    node->right->left=head;
  node->right=head;
  head->lower=value;
  node->upper=value;
  head->repr++;
  return(1);
}

int ispoint(seg)
SEGMENT *seg;
{
   if(seg->lower==seg->upper) return 1;
   return 0;
}

/* find the first node with the same class on the right of the node */
SEGMENT *find_same_class_on_right(seg,class)
SEGMENT *seg;
int class;
{
SEGMENT *s;
   if (seg !=NULL && seg->class==class)
       return(seg);
   for (s=seg; s !=NULL && ispoint(s); s=s->right)
       if (s->class==class)
          return(s);
   return(NULL);
}

/* find the first node with the same class on the left of the node */
SEGMENT *find_same_class_on_left(seg,class)
SEGMENT *seg;
int class;
{
  SEGMENT *s;

  if (seg !=NULL && seg->class==class)
    return(seg);
  for (s=seg; s !=NULL && ispoint(s); s=s->left) {
    if (s->class==class)
      return(s);
  }
  return(NULL);
}

void generalize(seg, f, value, class)
SEGMENT *seg;
FEATURE f;
VALUE value;
CLASS class;
{
  SEGMENT *leftseg, *rightseg;
  float disleft=INFINITE, disright=INFINITE;

  leftseg=find_same_class_on_left(seg->left,class);
  rightseg=find_same_class_on_right(seg->right,class);
  if (leftseg != NULL )		/* otherwise disleft= INFINITE */
    disleft=value - leftseg->upper;
  if (rightseg != NULL )	/* otherwise disright= INFINITE */
    disright=rightseg->lower - value;

  if (disleft > gen_limit[f] && disright > gen_limit[f]) {
    subdivide_range(seg,value,class);
    return;
  }
  if (disleft <=disright && 
      generalize_left(seg,leftseg,value)) return;

  if (disright<=gen_limit[f] &&  
      generalize_right(seg,rightseg,value) ) return;
}

void update_feature_partitioning (f, inst)
FEATURE f;
INSTANCE inst;
{
   SEGMENT *seg;

   seg = search (f, inst.value[f]); /* we know inst has a KNOWN value */

   if (seg->class == inst.class) 
     seg->repr++;
   else {  /* we need change the partitioning */
     if (seg->upper == inst.value[f]) 
       update_point (seg, inst.value[f], inst.class);
     else {
       if (seg->class == UNDETERMINED)
          generalize(seg, f, inst.value[f], inst.class);
       else 
          subdivide_range (seg, inst.value[f], inst.class);
     }
   }
}

int segment_count ()
{
  int f, sum = 0;
  SEGMENT *s;

  for (f=0; f < No_Features; f++) 
    for (s = Partitions[f]; s != NULL; s = s->right)
      sum++;
  return (sum);
}
